(* An implementation of bicycle mechanics from http://www.nbi.dk/~randlov/bicycle.ps.gz 
   using the reward shaping of:
   http://www.cs.duke.edu/~mgl/papers/icml03.ps.gz
   Consider this alpha quality.
*)
let c = 66. (* cm horizontal distance between where wheel touches ground and the center of mass *)
let dcm = 30. (* cm vertical distance from bicycle center of mass to rider center of mass *)
let h = 94. (* cm height of the center of mass *)
let l = 111. (* cm distance between front tire and back tire at ground level. *)
let mc = 15. (* kg mass of bicycle *)
let md = 1.7 (* kg mass of tire *)
let mp = 60. (* kg mass of rider *)
let r = 34. (* cm radius of tire *)
(*let v = 10. *. 1000. *. 100. /. 3600. *)(* cm/s velocity of bicycle *)
let v = 0.
let g = 10. *. 100. (* cm/s/s *)
let dt = 0.02 (* s *)

let m = mc +. mp (* kg *)
let phidot = v /. r (* /s *)
let idc = md *. r *. r (* kg*cm^2 *)
let idv = 3. /. 2. *. idc (* kg*cm^2 *)
let idl = 1. /. 2. *. idc (* kg*cm^2 *)
let ibnc = 13. /. 3. *. mc *. h ** 2. +. mp *. (h +. dcm) ** 2. (* kg cm^2 *)

let goal = (0., 10000.) 
(*let goal = (0., 0.)*) (* cm, cm *)
let goal_dist = 10. *. 100. (* cm *)

let sign x = if x > 0. then 1. else ~-. 1.

let safe x = 
  if x > 1. then 1. 
  else 
    if x < ~-. 1. 
    then ~-. 1.
    else x

let pi = 3.1415
let crashed (_,_,_,_,w,_,_,_,_) = abs_float (w /. pi) > 12. /. 180.

let update (oxb, oyb, oxf, oyf, ow, owdot, owdotdot, otheta, othetadot) (d, t) =
  let real_d = d +. Random.float 4. -. 2. in
  
  let rf = l /. abs_float (sin otheta) 
  and rb = l /. abs_float (tan otheta)
  and rcm = ((l -. c) ** 2. +. (l /. tan otheta) ** 2.) ** 0.5 
  and phi = ow +. atan (real_d /. h) in
  
  let thetadotdot = t -. idv *. phidot *. owdot /. idl in
  let w = ow +. owdot *. dt
  and theta = otheta +. othetadot *. dt 
  and wdot = owdot +. owdotdot *. dt 
  and thetadot = othetadot +. thetadotdot *. dt 
  and wdotdot = 1. /. ibnc *. (m *. g *. h *. sin phi -. cos phi *. (idc *. phidot *. othetadot +. sign otheta *. v ** 2. *. (md *. r /. rf +. md *. r /. rb +. m *. h /. rcm) )) in
  let psi = ~-. (atan2 (oxf -. oxb) (oyf -. oyb)) in
  let xf = oxf -. v *. dt *. sin (psi +. theta +. sign (psi +. theta) *. asin (safe (v *. dt /. 2. /. rf)))
  and yf = oyf +. v *. dt *. cos (psi +. theta +. sign (psi +. theta) *. asin (safe (v *. dt /. 2. /. rf)))
  and xb = oxb -. v *. dt *. sin (psi +. sign psi *. asin (safe (v *. dt /. 2. /. rb)))
  and yb = oyb +. v *. dt *. cos (psi +. sign psi *. asin (safe (v *. dt /. 2. /. rb))) in
  if classify_float w = FP_nan or classify_float wdot = FP_nan or classify_float wdotdot = FP_nan 
      or classify_float theta = FP_nan or classify_float thetadot = FP_nan 
      or classify_float xb = FP_nan 	or classify_float xb = FP_nan 	or classify_float xb = FP_nan
      or classify_float xb = FP_nan then begin
	output_string stderr 
	  ("bicycle: nan detected, ow = "^string_of_float ow^" and owdot = "^string_of_float owdot^
	   " and owdotdot = "
	   ^string_of_float owdotdot^" and otheta = "^string_of_float otheta^" and othetadot = "^string_of_float othetadot^
	   " and oxb = "^string_of_float oxb^
	   " and oyb = "^string_of_float oyb^" and oxf = "^string_of_float oxf^" and oyf = "^string_of_float oyf
	   ^" and real_d = "^string_of_float real_d^" and t = "^string_of_float t^"\n");
	flush stderr;
	exit 1;
      end
  else
    (xb, yb, xf, yf, w, wdot, wdotdot, theta, thetadot)
      
let reward ((new_xb, new_yb, new_xf, new_yf, new_w, _, _, _, _) as new_state) (xb, yb, xf, yf, _, _, _, _, _) =
  if new_xb <> nan & new_yb <> nan & new_xf <> nan & new_yf <> nan & xb <> nan & yb <> nan & xf <> nan & yf <> nan
  then
    let new_dist = ((new_xb -. fst goal) ** 2. +. (new_yb -. snd goal) ** 2. ) ** 0.5 
    and old_dist = ((xb -. fst goal) ** 2. +. (yb -. snd goal) ** 2.) ** 0.5 in
    if crashed new_state
    then 0. 
    else 
      if new_dist < goal_dist
      then 1.
      else 
	(old_dist -. new_dist) /. 1000.
  else 
    begin
      output_string stderr "nan detected";
      flush stderr;
      0.
    end
	  
let state_from_string s =
  try
    Scanf.sscanf s "%f %f %f %f %f %f %f %f %f"
      (fun xb yb xf yf w wdot wdotdot theta thetadot -> 
	(xb,yb,xf,yf,w,wdot,wdotdot,theta,thetadot))
  with (* the simulation broke, so place the bike in a crash position *)
    Scanf.Scan_failure s -> (0.,0.,0.,0.,pi,0.,0.,0.,0.) 


let action_from_string s = 
  Scanf.sscanf s "%f %f" (fun d t -> ((d -. 1.) *. 2.,(t -. 1.) *. 2.))

let project (x_1,y_1) (x_2,y_2) =
  if (x_1 <> 0. or y_1 <> 0.) & (x_2 <> 0. or y_2 <> 0.)
  then
    (x_1 *. x_2 +. y_1 *. y_2) 
      /. sqrt (x_1 *. x_1 +. y_1 *. y_1)
      /. sqrt (x_2 *. x_2 +. y_2 *. y_2) 
  else 0.

let offset (x_1,y_1) (x_2,y_2) =
  (x_1 -. x_2, y_1 -. y_2)

let orthogonal (x,y) = (y, ~-. x)

let obs_from_state (xb,yb,xf,yf,w,wdot,wdotdot,theta,thetadot) =
  let offset_goal = offset goal (xb,yb)
  and offset_front = offset (xf,yf) (xb,yb) in
  let normal_front = orthogonal offset_front in
  let proj = project offset_goal offset_front
  and ortho_project = project offset_goal normal_front in
  let angle_to_goal = atan2 ortho_project proj in
  
  if classify_float w = FP_nan or classify_float wdot = FP_nan or classify_float wdotdot = FP_nan 
      or classify_float theta = FP_nan or classify_float thetadot = FP_nan or classify_float angle_to_goal = FP_nan
  then begin
    output_string stderr 
      ("bicycle: nan detected, w = "^string_of_float w^" wdot = "^string_of_float wdot^" wdotdot = "
       ^string_of_float wdotdot^" theta = "^string_of_float thetadot^" angle_to_goal = "
       ^string_of_float angle_to_goal^" xb = "^string_of_float xb^" yb = "^string_of_float yb^" xf = "
       ^string_of_float xf^" yf = "^string_of_float yf^"\n");
    flush stderr;
  end;
  
  Printf.sprintf "%f %f %f %f %f %f" w wdot wdotdot theta thetadot angle_to_goal    

let string_from_state (xb,yb,xf,yf,w,wdot,wdotdot,theta,thetadot) =
  Printf.sprintf "%f %f %f %f %f %f %f %f %f" xb yb xf yf w wdot wdotdot theta thetadot

let rec make_initial _ = 
  let dist = sqrt (fst goal *. fst goal +. snd goal *. snd goal) in
  let x_b = Random.float (2. *. dist) -. dist
  and y_b = Random.float (2. *. dist)
  and x_f = Random.float 200. -. 100. 
  and y_f = Random.float 200. -. 100. in
  let norm = sqrt (x_f *. x_f +. y_f *. y_f) in
  let x_f = x_b +. 111. *. x_f /. norm 
  and y_f = y_b +. 111. *. y_f /. norm in
  (x_b, y_b, x_f, y_f)
    
let safe_read_line _ = 
  try
    read_line()
  with
    End_of_file -> exit 0

let update state action = 
  let state = state_from_string state in
  let action = action_from_string action in
  if crashed state 
  then ("","","")
  else
    let new_state = update state action in
    (obs_from_state new_state), (string_of_float (reward new_state state)), (string_from_state new_state)
      
let _ =
  output_string stdout "actions: {0,2}x{0,2}\n";
  flush stdout;
  let state = ref ""
  and seed = ref ""
  and action = ref "" in
  while true do
    begin 
      match !seed with
	"" -> ()
      | s -> Random.init (int_of_string s)
    end;
    let new_obs, new_reward, new_state =
      match !state with
      | "" -> 
	  let x_b,y_b,x_f,y_f = 0.,0.,111.,0. (*make_initial () *)in
	  let state_v = x_b,y_b,x_f,y_f,0.,0.,0.,0.,0. in
	  let state = string_from_state state_v
	  and obs = obs_from_state state_v in
	  (obs,"0.",state)
      | _ -> update !state !action in
    Printf.printf "%s\n%s\n%s\n" new_obs new_reward new_state;
    flush stdout;
    state := safe_read_line(); 
    seed := safe_read_line();
    action := safe_read_line()
  done;;
