(* An implementation of pole balancing mechanics from http://www.cs.duke.edu/~mgl/papers/icml03.ps.gz
   Consider this alpha quality. *)
let g = 9.8 (* m/s^2 *)
let m = 2.0 (* kg *)
let bigm = 8.0 (* kg *)
let l = 0.5 (* m *)
let alpha = 1. /. (m +. bigm)

let pi = 3.1415
let crashed theta = abs_float (theta /. pi) > 0.5
let dt = 0.02

let reward (theta, thetadot) =
  if crashed theta
  then 0.
  else 1.

let state_from_string s = Scanf.sscanf s "%f %f" (fun theta thetadot -> (theta,thetadot))

let action_from_string s = Scanf.sscanf s "%f" (fun force -> 50. *. (force -. 1.))
    
let obs_from_state (theta,thetadot) = Printf.sprintf "%f %f" theta thetadot

let string_from_state (theta,thetadot) = Printf.sprintf "%f %f" theta thetadot

let safe_read_line _ = 
  try
    read_line()
  with
    End_of_file -> exit 0

let update state action =
  let (theta,thetadot) = state_from_string state in
  let force = action_from_string action in
  if crashed theta
  then ("","","")
  else
    let real_force = force +. Random.float 20. -. 10. in
    let thetadotdot = (g *. sin theta -. alpha *. m *. l *. thetadot ** 2. *. sin (2. *. theta) /. 2. 
			 -. alpha *. cos theta *. real_force) 
	/. (4. *. l /. 3. -. alpha *. m *. l *. (cos theta) ** 2.) in
    let theta = theta +. thetadot *. dt 
    and thetadot = thetadot +. thetadotdot *. dt in
    (obs_from_state (theta,thetadot)), (string_of_float (reward (theta,thetadot))), (string_from_state (theta,thetadot))
      
let _ = 
  output_string stdout "actions: {0,2}\n";
  flush stdout;
  let state = ref ""
  and seed = ref ""
  and action = ref "" in
  while true do
    begin 
      match !seed with
	"" -> ()
      | s -> Random.init (int_of_string s)
    end;
    let (new_obs,new_reward,new_state) = 
      match !state with
      | "" -> 
	  let initial = "0 0"  in
	  (initial,"0.",initial)
      | _ -> update !state !action in
    Printf.printf "%s\n%s\n%s\n" new_obs new_reward new_state;
    flush stdout;
    state := safe_read_line();
    seed := safe_read_line();
    action := safe_read_line()
  done
