(* Code for a KD-Tree in caml *)

open List;;
open Kernel;;

(******* constants *****)
let tolerance = 0.01

(***** Helper functions *****)

(*for list of arrays*)
let fold_map = fun lfun initial list -> 
  let f = Myarray.map2 lfun in
  fold_left f initial list

(*** types *****)

(*type 'a trunk = Empty | FilledWith of 'a;;*)
type 'a point = {location : float array; weight : float; trunk : 'a option} ;;
type splitDim = (int * float);;
type bbox = ( (float array) * ( float array) );;
type 'a kdt = Node of splitDim * bbox * float * 'a kdt * 'a kdt 
  | Leaf of 'a point list ;;

(* Compute the bounding box of a list of double arrays*)
let boundingBox = fun pointList ->
  let len = Myarray.length (hd pointList) in
  let lowBound = 
    fold_map min (Myarray.make len infinity) pointList in
  let highBound = 
    fold_map max ( Myarray.make len neg_infinity) pointList in
  (lowBound, highBound)

(* Does the same for a list of points *)
let boundingBoxPoints = fun pointList ->
  let loc = fun point -> point.location in
  boundingBox (map loc pointList)
  
(* Could also do this for variance *)
(* Computes split direction and location *)
let splitDir = fun (lowBound, highBound) -> 
  let sizes = Myarray.map2 (-.) highBound lowBound in
  (*let len = length sizes in *)
  let maxindex = fun (ind1,x1) (ind2,x2)  -> 
    if x2 > x1 then 
      ( ind2, (x2 /. 2.0) +. (Myarray.get lowBound ind2))
    else
      (ind1, (x1 /. 2.0)  +. (Myarray.get lowBound ind1))
  in Myarray.fold_left_noinit_i maxindex sizes
  
let splitFilter =fun sd op point ->
  (op  (Myarray.get point.location (fst sd) ) ( (snd sd) ) ) 

let minBB = 0.0001;;

let rec insertTree = fun pointList ->
  let bb = boundingBoxPoints pointList in  
  let splitQ = (length pointList > 30) && 
    (distanceSquared (fst bb) (snd bb) > minBB ) in
  if splitQ then 
    let sd = splitDir bb in 
    let weights = map (fun p -> p.weight) pointList in  
    let alphaSum = fold_left (fun x1 x2 -> (abs_float x1) +. (abs_float x2) )
	0.0 weights in
    let (leftHalf,rightHalf) = partition (splitFilter sd (<) ) pointList in
    Node ((sd), bb, alphaSum, (insertTree leftHalf), (insertTree rightHalf) )
  else
    Leaf pointList
       

let corner = fun op x2 x1 ->
  op
    ( abs_float ( x1 -. (fst x2) ) )
    ( abs_float ( x1 -. (snd x2) ) )


(* Computing mimimum point in point centric coordinates *)
let minPoint = fun bb x ->
  let closestOp = fun diffHi diffLo ->
    if diffHi > 0.0 && diffLo > 0.0 then
      min diffHi diffLo
    else if diffHi < 0.0 && diffLo < 0.0 then
      max diffHi diffLo
    else
      0.0
  in
  let diffHi = Myarray.map2 (-.) (fst bb) x in
  let diffLo = Myarray.map2 (-.) (snd bb) x in
  Myarray.map2 closestOp diffHi diffLo



(* not sure that max and min contributions are correct *)
let rec retrieve = 
  fun tree center -> 
    match tree with
    | Node (sd,bb,alphaSum,l,r) -> 
	
	let bbsplit = Myarray.map2 (fun x y -> (x,y)) (fst bb) (snd bb) in
	
	(* Compute coordinates of nearest and furthest points to this
	   point in center-centric coordinates
	   Furthest point is invariably a corner. We can essentially
	   check corners....
	 *)
	let maxcorner = Myarray.map2 (corner max) bbsplit center    in 
	(* Closest point may be closer than any corner-- it might even be 0.0 *)
	let mincorner = minPoint bb center in

	(*compute their contribution *)
	let len = Array.length center in  
	let mincontrib = -1.0 *. alphaSum *. gaussKernel mincorner (Array.make len 0.0) in
	let maxcontrib =  1.0 *. alphaSum *. gaussKernel maxcorner (Array.make len 0.0) in

	let possibleError = maxcontrib -. mincontrib in 
	if ( possibleError < tolerance ) then begin
(*	  Printf.Printf "Tolerance Cut-Off: %f  %f \N" Maxcontrib Mincontrib;*)
	  let averageImpact = (maxcontrib +. mincontrib) /. 2.0  in 
	  averageImpact end
	else begin
(*	  Printf.printf "continuing down tree\n";*)
	  (retrieve l center) +. (retrieve r center) end
	    
    | Leaf pointList ->
	let locations = map (fun p -> p.location) pointList  in
	let weights = map (fun p -> p.weight) pointList  in begin
	  (*Printf.printf "At Leaf\n";*)
	  kernelFun locations weights center
	end

;;

(*let fakepoint = { location = [|0.2   ; 3.0|] ; weight = 1.0 ; *)
(*  trunk = FilledWith 1.0 };;*)

let 
(pointList: (float point) list) 
= [ 
{location = [|0.2   ; 3.0|] ; weight = 1.0 ; trunk = None} ;
{location = [|-0.25 ; 3.0|] ; weight = 1.0 ; trunk = None};
{location = [|-0.21 ; 3.2|] ; weight = 1.0 ; trunk = None};
{location = [|-0.21 ; 3.2|] ; weight = 1.0 ; trunk = None};
{location = [|-0.21 ; 3.2|] ; weight = 1.0 ; trunk = None};
{location = [|-0.21 ; 3.2|] ; weight = 1.0 ; trunk = None};
{location = [|-0.21 ; 3.2|] ; weight = 1.0 ; trunk = None};
{location = [|-0.27 ; 3.3|] ; weight = 1.0 ; trunk = None};
{location = [|-0.21 ; 3.2|] ; weight = 1.0 ; trunk = None};
{location = [|0.45 ; 3.21|] ; weight = 1.0 ; trunk = None}
];;

let center =   [|-0.21 ; 3.2|];;
(*let center =   [|0.45 ; 3.21|];;*)

let locations = map (fun p -> p.location) pointList;;  
let weights = map (fun p -> p.weight) pointList  ;;
kernelFun locations weights center;;


let t = insertTree(pointList);;

retrieve t  center;;
